#include "phc.h"
#include "frameBuffer.h"
#include "m33.h"
#include <math.h>
#include <float.h>
#include <fstream.h>

PHC::PHC(int _w, int _h, float hfov) :
 a(1.0f, 0.0f, 0.0f), b(0.0f, -1.0f, 0.0f), C(0.0f, 0.0f, 0.0f) ,
 w(_w), h(_h) {

  c[0] = -(float)w/2.0f;
  c[1] = (float)h/2.0f;
  float dhfov = hfov * 3.1415926535f / 180.0f;
  c[2] = c[0] / tanf(dhfov/2.0f);
  UpdateCamMatI();

}

void PHC::UpdateCamMatI() {

  camMatI = GetCamMat().Inverted();

}

M33 PHC::GetCamMat() {

  M33 ret;
  ret.SetColumn(0, a);
  ret.SetColumn(1, b);
  ret.SetColumn(2, c);
  return ret;

}

bool PHC::Project(V3 P, V3 &projP) {

  V3 q = camMatI*(P-C);
  if (q[2] <= 0.001f) {
    projP[2] = FLT_MAX;
    return false;
  }

  projP = V3(q[0]/q[2], q[1]/q[2], 1.0f/q[2]);
  if (
    projP[0] < 0.0f || projP[0] >= (float)w ||
    projP[1] < 0.0f || projP[1] >= (float)h)
    return false;

  return true;

}


V3 PHC::VD() {

  return (a^b).Normalized();

}

float PHC::FocalLength() {

  return c*VD();

}

void PHC::SetFocalLength(float fp) {

  V3 df = VD()*(fp-FocalLength());
  c = c+df;
  UpdateCamMatI();

}

void PHC::IncreaseFocalLength(float fls) {


  float f = FocalLength();
  float fp = f*fls;
  SetFocalLength(fp);
  UpdateCamMatI();

}

void PHC::ModifyResolution(int neww) {

  float scalef = (float)neww/(float)w;
  c = c * scalef;
  w = neww;
  h = (int)((float)h*scalef);
  UpdateCamMatI();

}

void PHC::TranslateForward(float tstep) {

  C = C + VD()*tstep;
  UpdateCamMatI();

}

void PHC::TranslateLeft(float tstep) {

  C = C + a.Normalized()*-tstep;
  UpdateCamMatI();

}


void PHC::LookAt(V3 &point, V3 &vd, float dist, V3 &quasiUp) {

  V3 Cp, ap, bp, cp;
  Cp = point - vd*dist;
  ap = (vd ^ quasiUp).Normalized() * a.Length();
  bp = (vd ^ ap).Normalized() * b.Length();
  cp = vd * FocalLength() - ap * (float) w * .5f - bp * (float) h *.5f;
  a = ap;
  b = bp;
  c = cp;
  C = Cp;
  UpdateCamMatI();

}

void PHC::Render(FrameBuffer *fb, PHC *phc, float f, unsigned int color) {

  float actualf = FocalLength();
  float scf = f/actualf;

  fb->DrawSegment3D(C, C+c*scf, *phc, color);
  fb->DrawSegment3D(C+c*scf, C+c*scf+a*(float)w*scf, *phc, color);
  fb->DrawSegment3D(C+c*scf+a*(float)w*scf, 
    C+c*scf+a*(float)w*scf+b*(float)h*scf, *phc, color);
  fb->DrawSegment3D(C+c*scf+a*(float)w*scf+b*(float)h*scf, 
    C+c*scf+b*(float)h*scf, *phc, color);
  fb->DrawSegment3D(C+c*scf+b*(float)h*scf, 
    C+c*scf, *phc, color);

  int gridStep = 40;
  for (int v = 0; v < h; v+=gridStep) {
    V3 pt0, pt1;
    pt0 = GetPoint(0.0f, (float)v);
    pt0 = C + (pt0-C)*scf;
    pt1 = GetPoint((float)w, (float)v);
    pt1 = C + (pt1-C)*scf;
    fb->DrawSegment3D(pt0, pt1, *phc, color);
  }
  for (int u = 0; u < w; u+=gridStep) {
    V3 pt0, pt1;
    pt0 = GetPoint((float)u, 0.0f);
    pt0 = C + (pt0-C)*scf;
    pt1 = GetPoint((float)u, (float)h);
    pt1 = C + (pt1-C)*scf;
    fb->DrawSegment3D(pt0, pt1, *phc, color);
  }

}

V3 PHC::GetPoint(float uf, float vf) {


  return C + c + a*uf + b*vf;

}

void PHC::Pan(float panAngle) {

  M33 csc, rot('y', panAngle);
  csc[0] = a.Normalized();
  csc[1] = b.Normalized()*-1;
  csc[2] = (b^a).Normalized();
  
  a = (csc.Inverted() * rot * csc) * a;
  b = (csc.Inverted() * rot * csc) * b;
  c = (csc.Inverted() * rot * csc) * c;
  UpdateCamMatI();

}

void PHC::Tilt(float tiltAngle) {

  M33 csc, rot('y', tiltAngle);
  csc[0] = b.Normalized();
  csc[1] = a.Normalized();
  csc[2] = (b^a).Normalized();
  
  a = (csc.Inverted() * rot * csc) * a;
  b = (csc.Inverted() * rot * csc) * b;
  c = (csc.Inverted() * rot * csc) * c;
  UpdateCamMatI();

}

void PHC::Roll(float rollAngle) {

  M33 csc, rot('y', rollAngle);
  csc[0] = b.Normalized();
  csc[1] = (b^a).Normalized();
  csc[2] = a.Normalized();
  
  a = (csc.Inverted() * rot * csc) * a;
  b = (csc.Inverted() * rot * csc) * b;
  c = (csc.Inverted() * rot * csc) * c;
  UpdateCamMatI();

}


void PHC::TranslateUp(float tstep) {

  C = C + b.Normalized()*-1.0f*tstep;
  UpdateCamMatI();

}

void PHC::Save(char *fname) {


  ofstream ofs(fname);
  if (ofs.ios::fail()) {
    cerr << "ERROR: could not open: " << fname << endl;
    return;
  }

  ofs << a << endl << b << endl << c << endl << C;
  ofs.close();

}

void PHC::Load(char *fname) {


  ifstream ifs(fname);
  if (ifs.ios::fail()) {
    cerr << "ERROR: could not open: " << fname << endl;
    return;
  }

  ifs >> a >> b >>  c >> C;
  ifs.close();

  UpdateCamMatI();

}
