#include "tmesh.h"
#include <fstream.h>
#include <float.h>
#include <strstrea.h>

TMesh::TMesh() {

  color = 0x00000000;
  vertsN = 0;
  verts = 0;
  trisN = 0;
  tris = 0;
  litVerts = 0;

}


void TMesh::Load(char *fileName) {

  ifstream ifs(fileName);
  if (ifs.ios::fail()) {
    cerr << "ERROR: cannot open: " << fileName << endl;
    return;
  }

  if (vertsN)
    delete verts;
  vertsN = 0;
  if (trisN)
    delete tris;
  trisN = 0;

  ifs >> vertsN;
  verts = new V3[vertsN];
  char lineBuffer[1000];
  ifs.getline(lineBuffer, 999);
  for (int vi = 0; vi < vertsN; vi++) {
    char lineBuffer[1000];
    ifs.getline(lineBuffer, 999);
    istrstream istr(lineBuffer);
    istr >> verts[vi];
  }
  cerr << "INFO: read " << vertsN << " vertices" << endl;


  ifs >> trisN;
  tris = new unsigned int[trisN*3];
  ifs.getline(lineBuffer, 999);
  for (int ti = 0; ti < trisN; ti++) {
    char lineBuffer[1000];
    ifs.getline(lineBuffer, 999);
    istrstream istr(lineBuffer);
    istr >> tris[3*ti+0] >> tris[3*ti+1] >> tris[3*ti+2];
  }

  cerr << "INFO: read " << trisN << " triangles" << endl;

  ifs.close();

  Allocate();
  ComputeNormals();

}

void TMesh::Allocate() {

  litVerts = new V3[vertsN];
  normals = new V3[vertsN];

}

bool TMesh::BackFacing(int tri, PHC *phc) {

  V3 v0 = verts[tris[3*tri+0]];
  V3 v1 = verts[tris[3*tri+1]];
  V3 v2 = verts[tris[3*tri+2]];
  V3 normal = ((v1-v0)^(v2-v0)).Normalized();
  if (normal*((phc->GetC()-v0).Normalized()) < 0.0f)
    return true;
  return false;

}

void TMesh::RenderFM(FrameBuffer *fb, PHC *phc) {


  Light(phc->VD()*-1.0f);

  V3 *projVerts = new V3[vertsN];
  for (int vi = 0; vi < vertsN; vi++) {
    phc->Project(verts[vi], projVerts[vi]);
  }

  for (int tri = 0; tri < trisN; tri++) {

    if (BackFacing(tri, phc)) {
      continue;
    }

    // projection
    V3 projVs[3];
    V3 triVerts[3];
    V3 vertColors[3];
    bool triangleOK = true;
    for (int vi = 0; vi < 3; vi++) {
      triVerts[vi] = verts[tris[3*tri+vi]];
      projVs[vi] = projVerts[tris[3*tri+vi]];
      vertColors[vi] = litVerts[tris[3*tri+vi]];
      if (projVs[vi][0] == FLT_MAX) {
        triangleOK = false;
        break;
      }
    }
    if (!triangleOK)
      continue;

    // find axes aligned bounding box
    BBox bbox(projVs[0]);
    bbox.AddPoint(projVs[1]);
    bbox.AddPoint(projVs[2]);
    
    int left = (int) (bbox.corners[0][0] + 0.5f);
    int right = (int) (bbox.corners[1][0] - 0.5f);
    int top = (int) (bbox.corners[0][1] + 0.5f);
    int bottom = (int) (bbox.corners[1][1] - 0.5f);

    if (left < 0)
      left = 0;
    if (right < 0)
      continue;
    if (top < 0)
      top = 0;
    if (bottom < 0)
      continue;
    if (left > fb->w-1)
      continue;
    if (right > fb->w-1)
      right = fb->w-1;
    if (bottom > fb->h-1)
      bottom = fb->h-1;
    if (top > fb->h-1)
      continue;


    // rasterization setup
    V3 eeqs[3];
    for (int ei = 0; ei < 3; ei++) {
      int _ei = (ei+1)%3;
      eeqs[ei][0] = projVs[_ei][1]-projVs[ei][1];
      eeqs[ei][1] = projVs[ei][0]-projVs[_ei][0];
      eeqs[ei][2] = -projVs[ei][1]*eeqs[ei][1] - projVs[ei][0]*eeqs[ei][0];
      int __ei = (_ei+1)%3;
      V3 P2 = projVs[__ei];
      P2[2] = 1.0f;
      if (eeqs[ei]*P2 < 0.0f) {
        eeqs[ei] = eeqs[ei]*-1.0f;
      }
    }

    M33 ssim;
    ssim[0] = V3(projVs[0][0], projVs[0][1], 1.0f);
    ssim[1] = V3(projVs[1][0], projVs[1][1], 1.0f);
    ssim[2] = V3(projVs[2][0], projVs[2][1], 1.0f);
    ssim = ssim.Inverted();
    M33 colorsM;
    colorsM[0] = vertColors[0];
    colorsM[1] = vertColors[1];
    colorsM[2] = vertColors[2];
    M33 colorABCs = ssim*colorsM;
    colorABCs = colorABCs.Transposed();
    V3 zABC = ssim*V3(projVs[0][2], projVs[1][2], projVs[2][2]);

    M33 vcm;
    vcm.SetColumn(0, triVerts[0]-phc->GetC()); 
    vcm.SetColumn(1, triVerts[1]-phc->GetC()); 
    vcm.SetColumn(2, triVerts[2]-phc->GetC()); 
    M33 qmsi = vcm.Inverted() * phc->GetCamMat();
    V3 denABC(
      qmsi[0][0]+qmsi[1][0]+qmsi[2][0],
      qmsi[0][1]+qmsi[1][1]+qmsi[2][1],
      qmsi[0][2]+qmsi[1][2]+qmsi[2][2]
      );
    V3 redv(vertColors[0][0], vertColors[1][0], vertColors[2][0]);
    V3 greenv(vertColors[0][1], vertColors[1][1], vertColors[2][1]);
    V3 bluev(vertColors[0][2], vertColors[1][2], vertColors[2][2]);
    V3 numRedABC(qmsi.GetColumn(0)*redv, 
      qmsi.GetColumn(1)*redv, 
      qmsi.GetColumn(2)*redv);
    V3 numGreenABC(qmsi.GetColumn(0)*greenv, 
      qmsi.GetColumn(1)*greenv, 
      qmsi.GetColumn(2)*greenv);
    V3 numBlueABC(qmsi.GetColumn(0)*bluev, 
      qmsi.GetColumn(1)*bluev, 
      qmsi.GetColumn(2)*bluev);

    // rasterization loops
    for (int v = top; v <= bottom; v++) {
      for (int u = left; u <= right; u++) {
        V3 pixPoint(0.5f+(float)u, 0.5f+(float)v, 1.0f);
        if (
          eeqs[0]*pixPoint < 0 ||
          eeqs[1]*pixPoint < 0 ||
          eeqs[2]*pixPoint < 0) {
          continue;
        }
        // we are inside triangle
        float currz = zABC*pixPoint;
        if (currz < fb->ZB(u, v)) {
          continue;
        }
        fb->ZB(u, v) = currz;                
        // we are visible        

        V3 currColor;

        currColor[0] = (numRedABC*pixPoint) / (denABC*pixPoint);
        currColor[1] = (numGreenABC*pixPoint) / (denABC*pixPoint);
        currColor[2] = (numBlueABC*pixPoint) / (denABC*pixPoint);

        currColor = colorABCs * pixPoint;
        
        fb->Set(u, v, currColor.ToUInt());

      }
    }

  }

  delete projVerts;
  

}

void TMesh::RenderWF(FrameBuffer *fb, PHC *phc) {

  for (int tri = 0; tri < trisN; tri++) {
    for (int vi = 0; vi < 3; vi++) {
      fb->DrawSegment3D(verts[tris[3*tri+vi]], verts[tris[3*tri+((vi+1)%3)]], 
        *phc, color);
    }
  }

}


V3 TMesh::FindCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int i = 0; i < vertsN; i++) {
    ret = ret + verts[i];
  }
  return ret / (float) vertsN;

}


void TMesh::ScaleToMaxDim(float maxDim) {

  int vi;
  BBox bbox(verts[0]);
  for (vi = 0; vi < vertsN; vi++) {
    bbox.AddPoint(verts[vi]);
  }

  float oldMaxDim = bbox.GetDiagonal();

  float scaleFactor = maxDim/oldMaxDim;

  V3 center = FindCenter();
  TranslateBy(center);
  ScaleBy(scaleFactor);
  TranslateBy(center*-1.0f);
  
}


void TMesh::TranslateBy(V3 translationVector) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] + translationVector;
  }

}

void TMesh::TranslateTo(V3 newCenter) {

  V3 center = FindCenter();
  TranslateBy(newCenter-center);

}

void TMesh::ScaleBy(float scf) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] * scf;
  }

}

void TMesh::ComputeNormals() {

  int i;
  for (i = 0; i < vertsN; i++) {
    normals[i] = V3(0.0f, 0.0f, 0.0f);
  }
  for (i = 0; i < trisN; i++) {
    V3 n = 
      ((verts[tris[3*i+1]]-verts[tris[3*i+0]]) ^
      (verts[tris[3*i+2]]-verts[tris[3*i+0]])).Normalized();
    normals[tris[3*i+0]] = normals[tris[3*i+0]] + n;
    normals[tris[3*i+1]] = normals[tris[3*i+1]] + n;
    normals[tris[3*i+2]] = normals[tris[3*i+2]] + n;
  }
  for (i = 0; i < vertsN; i++) {
    normals[i] = normals[i].Normalized();
  }

}

void TMesh::Light(V3 lightVector) {

  float ka = 0.4f;
  V3 baseColor;
  baseColor.SetFromUint(color);
  for (int i = 0; i < vertsN; i++) {
    float kd = lightVector*normals[i];
    if (kd < 0.0f) kd = 0.0f;
    litVerts[i] = baseColor*(ka+(1.0f-ka)*kd);
  }

  if (vertsN == 4) {
    litVerts[0] = V3(255.0f, 0.0f, 0.0f);
    litVerts[3] = V3(255.0f, 0.0f, 0.0f);
    litVerts[1] = V3(0.0f, 0.0f, 0.0f);
    litVerts[2] = V3(0.0f, 0.0f, 0.0f);
  }

}

float TMesh::FindDiagonal() {

  int vi;
  BBox bbox(verts[0]);
  for (vi = 0; vi < vertsN; vi++) {
    bbox.AddPoint(verts[vi]);
  }

  return bbox.GetDiagonal();

}

TMesh::TMesh(char shape) {

  color = 0xFF0000FF;

  switch (shape) {
  case 'q': {
    vertsN = 4;
    trisN = 2;
    verts = new V3[vertsN];
    tris = new unsigned int[trisN*3];
    Allocate();
    verts[0] = V3(-20.0f, -2.0f, 0.0f);
    verts[1] = V3(+20.0f, -2.0f, 0.0f);
    verts[2] = V3(+20.0f, +2.0f, 0.0f);
    verts[3] = V3(-20.0f, +2.0f, 0.0f);
    tris[0] = 0;
    tris[1] = 1;
    tris[2] = 2;
    tris[3] = 2;
    tris[4] = 3;
    tris[5] = 0;
    break;
  }
  default: {
    cerr << "ERROR: do not know shape " << shape << endl;
  }
  }

  ComputeNormals();

}